// Open Laser Show DAC
// Copyright (c) 2011 Chris Favreau
// cfavreau@fab-favreau.com

#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/pgmspace.h>     // Macros to allow data storage in program (flash) space
#include "standard.h"
#include "usart.h"
#include "spi.h"
#include "timer.h"
#include "switch.h"


// ILDA 25 Pin D-Sub MAP
// Pin     Function
// -----------------------------------------
//  1      X+
//  2      Y+
//  3      NC
//  4      Interlock (Connected to Pin 17)
//  5      R+
//  6      G+
//  7      B+
//  8      User1+ (Harp Pos. Bit 0)
//  9      User2+ (Harp Pos. Bit 1)
// 10      User3+ (Harp Pos. Bit 2)
// 11      User4+ (Harp Pos. Bit 3)
// 12      NC
// 13      Shutter
// 14      X-
// 15      Y-
// 16      NC
// 17      Interlock (Connected to Pin 4)
// 18-20   Ground (for color outputs)
// 21      User1- (Harp Enable Input - 0 = Enable - See Code below to enable this feature)
// 22      User2- (Harp Sync/Clk - Lo to Hi means the harp has reached position X indicated by the Harp Pos. Bits 0 to 15)
// 23      NC (User3-)
// 24      NC (User4-)
// 25      Ground

// Arduino/AVR Port Map
// Jmp   Arduino Pin	AVR Pin		Direction	OLSD Func
//-----------------------------------------------------------------------------
// J3    Digital 13     PB4 (SCK)       Output          SPI Clock to DAC
// J3    Digital 11     PB3 (MOSI)      Output          SPI Data to DAC
// J3    Digital 10     PB2 (SS)        Output          #CS DAC
// J3    Digital 9      PB1             Output          #LDAC DAC
// J3    Digital 8      PB0             Output          #CTS on FTI Chip (Optional) *** Not needed for laser harp
// J1    Digital 7      PD7             Output          ILDA Pin 11 (User4+) - Harp Pos. Bit 3
// J1    Digital 6      PD6             Output          ILDA Pin 10 (User3+) - Harp Pos. Bit 2
// J1    Digital 5      PD5             Output          ILDA Pin 8  (User2+) - Harp Pos. Bit 1
// J1    Digital 4      PD4             Output          ILDA Pin 7  (User1+) - Harp Pos. Bit 0
// J1    Digital 3      PD3             Output          ILDA Pin 22 (User2-) - Harp Pos. Sync/Clk (sync on positive edge - lo to hi)
// J1    Digital 2      PD2             Input           ILDA Pin 21 (User1-) - Harp Output Enable (0 = Enable - safety feature?)
// J2    Analog 0       PC0             Output          Red TTL
// J2    Analog 1       PC1             Output          Green TTL
// J2    Analog 2       PC2             Output          Blue TTL
// J2    Analog 3       PC3             Output          Shutter
// J2    Analog 4       PC4             Output          Status LED (1 == ON)
// J2    Analog 5       PC5             Input           User Switch (0 == Pressed)

// Important Constants for defining the output of the laser
// Please adjust only the NUMBER_OF_STRINGS

#define ENABLE_ENABLE          FALSE   // Use TRUE to ENABLE the Safety Enable or FALSE to disable it
#define MAX_NUMBER_OF_STRINGS  16      // We only have 4 position bits for the strings so the max is 16
#define NUMBER_OF_STRINGS      1       // Preferably an even number (8 == 1 octave) MAX = 16
#define DAC_WIDTH              4096    // DO NOT CHANGE - 12 bit DAC
#define STEP_SIZE              32      // DAC step size ... do not make any jumps bigger than this
#define POINT_DELAY            16      // Number of points to delay for a string
#define OUTPUT_RATE            10000   // 10kpps output rate by default

// Store the Identifier in FLASH
unsigned char OLSD_IDENTIFIER_STRING[] PROGMEM = "OLSD_LASER_HARP_V1.0.0";

_WORD spi_word = 0;

// Pointer to the reset vector for the firmware
void (*reset)( void ) = 0x0000;
void blank_and_center(void);

#define LED_ON  portc.bit4 = 1
#define LED_OFF portc.bit4 = 0

// ----------------------------------------------------------------------

inline void NoteOn(_BYTE cmd, _BYTE pitch, _BYTE velocity)
{
  USART_Write(cmd);
  USART_Write(pitch);
  USART_Write(velocity);
}

volatile _BYTE StringToNote[16] = {0x3C, 0x3E, 0x40, 0x41, 0x43, 0x45, 0x47, 0x48, 0x4A, 0x4C, 0x4D, 0x4F, 0x51, 0x53, 0x54, 0x56};
volatile _BYTE StringOn[16];

void ResetStrings(void)
{
  _BYTE i;
  for (i = 0; i < 16; i++) StringOn[i] = 0;
}

void PlayString(_BYTE nString)
{
  // Check to see if the string is on already
  // If it is then don't send the note... 
  // Just set the StringOn count for the string
  if (!IsStringOn(nString))
  {
    NoteOn(0x90, StringToNote[nString], 0x45);
  }
  //NoteOn(0x90, 0x3C + nString, 0x45);
  StringOn[nString] = 60;
}

_BOOL IsStringOn(_BYTE nString)
{
  return (StringOn[nString] > 0);
}

void StopString(_BYTE nString)
{
  NoteOn(0x90, StringToNote[nString], 0x00);
  StringOn[nString] = FALSE;
}

inline void PrintIdentifier(void)
{
	// Get the OLSD Identifier from Flash instead of RAM
	// so we don't use up valuable RAM
	unsigned char i = 0;
	unsigned char byte = pgm_read_byte(&OLSD_IDENTIFIER_STRING[i]);
	// Loop until we run into the NULL terminator
	while (byte)
	{
		USART_Write(byte);
		byte = pgm_read_byte(&OLSD_IDENTIFIER_STRING[i++]);
	}
}

void blank_and_center(void)
{
	// Define a center
	#define CENTER_X 2047
	#define CENTER_Y 2047

	portb.bit1 = 1;

	// Load DAC A
	// X X X PowerMode Load Output AddrB AddrA 
	// pull down chip select!
	spi_word = 0b0111000000000000 | CENTER_X;
	portb.bit2 = 0;
	SPDR = spi_word >> 8;
	while (!(SPSR & (1 << SPIF)));
	SPDR = spi_word & 0x00FF;
	while (!(SPSR & (1 << SPIF)));
	// pull up chip select!
	portb.bit2 = 1;
	asm("nop");
	// NOP NOP NOP NOP!??? how many?
	// Load DAC B and update the output (assign a 0 to output)
	// pull down chip select!
	spi_word = 0b1111000000000000 | CENTER_Y;
	portb.bit2 = 0;
	SPDR = spi_word >> 8;
	while (!(SPSR & (1 << SPIF)));
	SPDR = spi_word & 0x00FF;
	while (!(SPSR & (1 << SPIF)));
	// Pull up chip select!
	portb.bit2 = 1;
	portb.bit1 = 0;

        // Red = 0
        // Green = 0
        // Blue = 0
        //PORTC = (PORTC & 0xF8);
        PORTD = 0b00000011;
        PORTC = 0b00110000;

        // Shutter = 1
        // LED - BIT4 - leave ALONE! 
        // Switch - BIT5 - leave ALONE!
        portc.bit3 = 1;
}

inline void output_point(_WORD x, _WORD y, _BYTE r, _BYTE g, _BYTE b)
{
  // Bounds check X and Y
  if (x > 4095) x = 4095;
  if (y > 4095) y = 4095;
  
  // Pull the LDAC line high
  portb.bit1 = 1;
  
  // Load DAC A
  // X X X PowerMode Load Output AddrB AddrA 
  spi_word = 0b0111000000000000 | (y & 0x0FFF);
  // Pull down chip select so the DAC knows we are taling to it
  portb.bit2 = 0;
  // Upper Byte of the command
  SPDR = spi_word >> 8;
  while (!(SPSR & (1 << SPIF)));
  // Lower Byte of the command
  SPDR = spi_word & 0x00FF;
  while (!(SPSR & (1 << SPIF)));
  // Pull up chip select
  portb.bit2 = 1;
  
  // Load DAC B and update the output (assign a 0 to output)
  spi_word = 0b1111000000000000 | (x & 0x0FFF);
  // Pull down chip select so the DAC knows we are taling to it
  portb.bit2 = 0;
  // Upper Byte of the command
  SPDR = spi_word >> 8;
  while (!(SPSR & (1 << SPIF)));
  // Lower Byte of the command
  SPDR = spi_word & 0x00FF;
  while (!(SPSR & (1 << SPIF)));
  // Pull up chip select
  portb.bit2 = 1;
  
  // Write the R G B bits
  /*
  _BYTE rgb = PORTC & 0b00111000;
  if (r > 0) rgb |= 0x01;
  if (g > 0) rgb |= 0x02;
  if (b > 0) rgb |= 0x04;
  PORTC = rgb;
  */
  
  // Convert colors to 3 bit
  r = r >> 5;
  b = b >> 5;
  g = g >> 5;
  
  PORTD = 0b00000011 | (r << 5) | (g << 2);
  PORTC = 0b00110000 | b;

  // Open the shutter
  // TODO
  
  // Pull the LDAC line LOW to load the DAC's values
  portb.bit1 = 0;
}

void init_ports(void)
{
  // Initialize our IO Ports
  // Port C = Button + LED + Shutter + B
  DDRC = 0b00011111;
  PORTC = 0b00111000;

  // Light Gun Trigger on PC3
  ddrc.bit3 = 0;
  portc.bit3 = 1;
  // Light Gun Light on PC5
  ddrc.bit5 = 0;
  portc.bit5 = 1;
  
  // Port D - Serial + RG
  DDRD = 0b11111100;
  PORTD = 0x00;
  
  // Initialize our SPI port to output to our DAC
  ddrb.bit2 = 1;
  portb.bit2 = 1;
  SPI_Init();

  // LDAC - disable
  ddrb.bit1 = 1;
  portb.bit1 = 1;

  // CTS Pin
  ddrb.bit0 = 1;
  portb.bit0 = 0;

  // Status LED
  ddrc.bit4 = 1;
  portc.bit4 = 1;		// Turn the LED ON
}

// Status LED Functions
_WORD status_counter = 0;
void do_led()
{
  // Do the Status LED
  status_counter++;
  // On or Off?
  if (status_counter < 500)
  {
    //LED_ON;
  }
  else
  {
    //LED_OFF;
  }
  // Reset the counter if it is time
  if (status_counter > 1000) status_counter = 0;
}

void blankoutput(void)
{
  // BLANK!!!!
  // Red = 0
  // Green = 0
  // Blue = 0
  //PORTC = (PORTC & 0xF8);
  PORTD = 0b00000011;
  PORTC = 0b00110000;

  // Shutter = 1
  // LED - BIT4 - leave ALONE! 
  // Switch - BIT5 - leave ALONE!
  portc.bit3 = 1;
}

void outputblankpoint(int x)
{
  // Wait for out output flag to change
  while (!output_flag) asm("nop");
  
  // Reset our output flag/counter
  output_flag = 0;

  output_point(x, 2047, 0, 0, 0);

  LED_OFF;
  //do_led();
}

void outputlitpoint(int x)
{
  // Wait for out output flag to change
  while (!output_flag) asm("nop");
  
  // Reset our output flag/counter
  output_flag = 0;

  output_point(x, 2047, 255, 255, 255);

  LED_ON;
  //do_led();
}

void drawto(int xfrom, int xto)
{
  for (int x = xfrom; x < xto; x += STEP_SIZE)
  {
    outputblankpoint(x);
  }
  outputblankpoint(xto);
}

void drawstring(int x, _BYTE nString)
{
  outputblankpoint(x);
  outputblankpoint(x);
  outputblankpoint(x);
  outputblankpoint(x);
  
  outputlitpoint(x);
  
  // Set string bits
  // nString should only be from 0 to 15
  //PORTD = (PORTD & 0x0F) | (nString << 4);
  
  // Toggle sync line ON
  portd.bit3 = 1;
  
  _BYTE nDelay = 0;
  while (nDelay < (POINT_DELAY - 1))
  {
      outputlitpoint(x);
      outputblankpoint(x);
      // Increment the delay counter
      nDelay++;
  }
  
  // Check to see if the string triggered our light sensor
  if (switch_pressed > 0)
  {
    //NoteOn(0x90, 0x3C + nString, 0x45);
    PlayString(nString);
    //USART_Write('a' + nString);
  }
  switch_pressed = 0;
  // Notes go off in the switch.cpp for now
  
  // Toggle the sync line OFF
  //portd.bit3 = 0;
  
  outputblankpoint(x);
}

void retrace(int x)
{
  while (x > 0)
  {
    outputblankpoint(x);
    x -= STEP_SIZE;
  }
}

int main(void)
{
  // Initialize our port IO
  init_ports();
	
  // Initialize our USART at 2 Mbps
  USART_Init();
  
  // Initialize the output point timer
  init_timer();

  // Blank and center
  blank_and_center();

  // Reset our strings
  ResetStrings();
  
  // Initialize the button input device
  init_switch();
  
  // Turn on interrupts
  sei();

  // Print Boot Message
  PrintIdentifier();
  
  // Set the output rate
  set_timer_pps(30000);
  
  // Set the default number of strings
  _BYTE nStrings = NUMBER_OF_STRINGS;
  // Bounds check it
  if (nStrings > MAX_NUMBER_OF_STRINGS) nStrings = MAX_NUMBER_OF_STRINGS;
  
  // Loop forever
  int i = 0;
  int previ;
  int iStepSize = (DAC_WIDTH - 1) / (nStrings - 1);
  _BYTE iString = 0;
  
  while (1)
  {
    // Make sure the output is enabled first!
    if (ENABLE_ENABLE && (pind.bit2 != 0))
    {
      // Blank the colors
      blankoutput();
      // Continue the loop
      continue;
    }
    /*
    // Change the number of strings if the user presses the button!
    if (switch_pressed == 1)
    {
      switch_pressed = 0;
      nStrings--;
      // We cannot display less than 2 strings...
      // 1 string is a special case...
      // If you want 1 string... then we need to just set a flag
      // and handle it...
      if (nStrings < 1) nStrings = MAX_NUMBER_OF_STRINGS;
      if (nStrings > 1)
        iStepSize = (DAC_WIDTH - 1) / (nStrings - 1);
      else
        iStepSize = (DAC_WIDTH - 1);      // This is not important as we aren't going to use it
    }
    */
    
    if (nStrings < 2)
    {
      // Draw a single string in the middle
      drawstring(2047, 0);
      // Continue the main loop
      continue;
      // *** We never should go into the other display loop with 1 string
      // *** although the step size calculation of (DAC_WIDTH - 1) will display 2 strings in this case
    }
    
    i = 0;
    previ = i;
    iString = 0;
    while (i < DAC_WIDTH)
    {
      // Move the mirror to the string position
      drawto(previ, i);
      
      // Output the sring
      drawstring(i, iString);
      
      // Next position
      previ = i;
      i += iStepSize;
      
      // Next string
      iString++;
    }
    // Reset i position so we do NOT overflow and cause the scanners to go out of bounds
    i = (DAC_WIDTH - 1);
    retrace(i);
  }
  
  // We should never get here...
  return 1;
}
